class Scan<Data, Tally, Result> {
  int n, t, t_plus, single_len, last_t, last_len;
  Data[] array;
  Result[] result;
  FullEmpty<Tally>[] nodevalVar;
  FullEmpty<Tally>[] pvalVar;
  FullEmpty<Result> resultVar;
  
  public interface Server<Data, Tally, Result> {
    public Tally init();
    public Tally accum(Tally t, int i, Data[] d);
    public Tally combine(Tally t1, Tally t2);
    public Tally scanGen(Tally t, int i, Data[] d, Result[] r);
  }

  public interface ServerMaker<Data, Tally, Result> {
    public Server<Data, Tally, Result> make();
  }

  public interface Client<Data, Result> {
    public int getN();
    public Data[] getArray();
    public Result[] getResult();
    
    public void setup();
    public void printResult();
  }

  class ServerThread extends Thread {
    Server<Data, Tally, Result> s;
    int index, offset, size;

    public ServerThread(Server<Data, Tally, Result> s, int index, int offset, int size) {
      this.s = s;
      this.index = index;
      this.offset = offset;
      this.size = size;
    }

    public void run() {
      Tally tally;
      
      tally = localTally();
      
      for (int stride = 1; stride < t_plus; stride *= 2) {
        if (index % (2 * stride) == 0) {
          if (index + stride < t) {
            pvalVar[index + stride].put(tally);
            tally = s.combine(tally, nodevalVar[index + stride].get());
          }
        } else {
          nodevalVar[index].put(tally);
          break;
        }
      }

      localScanGen(propagateParentTally());
    }

    Tally localTally() {
      Tally tally = s.init();

      for (int i = 0; i < size; i++) {
        tally = s.accum(tally, i + offset, array);
      }

      return tally;
    }

    Tally propagateParentTally() {
      Tally ptally;
  
      if (index == 0)
        ptally = s.init();
      else {
        /* Parent supplies left total then parent */
        ptally = s.combine(pvalVar[index].get(), pvalVar[index].get());
      }
  
      /* Propagate total from parent to right children */
      for (int stride = 1; stride < t_plus; stride *= 2) {
        if (index % (2 * stride) == 0) {
          if (index + stride < t)
            pvalVar[index + stride].put(ptally);
        } else
          break;
      }
      
      return ptally;
    }

    void localScanGen(Tally ptally) {
      for (int i = 0; i < size; i++)
        ptally = s.scanGen(ptally, i + offset, array, result);
    }
  }

  public void run(String[] argv, ServerMaker<Data, Tally, Result> sm, Client<Data, Result> c) {
    Thread s[];

    c.setup();
    n = c.getN();
    array = c.getArray();
    result = c.getResult();

    t = Integer.valueOf(argv[0]);
    for (t_plus = 1; t_plus < t; t_plus *= 2) { }

    resultVar = new FullEmpty<Result>();
    nodevalVar = new FullEmpty[t];
    pvalVar = new FullEmpty[t];
    for (int i = 0; i < t; i++) {
      nodevalVar[i] = new FullEmpty<Tally>();
      pvalVar[i] = new FullEmpty<Tally>();
    }

    single_len = n / t;
    
    /* Last server waiting_gets leftovers if the array doesn't divide evenly: */
    last_t = t - 1;
    last_len = n - ((t - 1) * single_len);

    s = new Thread[t];
    for (int i = 0; i < t; i++) {
      s[i] = new ServerThread(sm.make(), i, i * single_len, ((i == last_t) ? last_len : single_len));
      s[i].start();
    }

    for (int i = 0; i < t; i++) {
      try {
        s[i].join();
      } catch (InterruptedException e) {
        System.exit(0);
      }
    }
    
    c.printResult();
  }
}
