#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "fullempty.h"

struct fullempty_t {
  int full;
  pthread_mutex_t m;
  pthread_cond_t full_c, empty_c;
  int size;
  char data[1];
};

struct fullempty_t *fullempty_create(int size)
{
  struct fullempty_t *var;

  var = (struct fullempty_t *)malloc(sizeof(struct fullempty_t) + size);

  var->full = 0;
  pthread_mutex_init(&var->m, NULL);
  pthread_cond_init(&var->full_c, NULL);
  pthread_cond_init(&var->empty_c, NULL);
  var->size = size;

  return var;
}

void fullempty_put(struct fullempty_t *var, int size, void *data)
{
  if (size != var->size) {
    printf("fullempty data size mismatch\n");
    abort();
  }

  pthread_mutex_lock(&var->m);
  while (var->full) {
    pthread_cond_wait(&var->empty_c, &var->m);
  }
  var->full = 1;
  memcpy(var->data, data, size);
  pthread_cond_signal(&var->full_c);
  pthread_mutex_unlock(&var->m);
}

void fullempty_get(struct fullempty_t *var, int size, void *data)
{
  if (size != var->size) {
    printf("fullempty data size mismatch\n");
    abort();
  }

  pthread_mutex_lock(&var->m);
  while (!var->full) {
    pthread_cond_wait(&var->full_c, &var->m);
  }
  var->full = 0;
  memcpy(data, var->data, size);
  pthread_cond_signal(&var->empty_c);
  pthread_mutex_unlock(&var->m);
}
