#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "config.h"
#include "scan_client.h"
#include "scan_server.h"
#include "fullempty.h"

static int t, t_plus, single_len, last_t, last_len;

static struct fullempty_t **nodeval_var;
static struct fullempty_t **pval_var;

static void *start_server(void *);

int main(int argc, char **argv)
{
  pthread_t *c;
  int i;

  if (argc != 2) {
    printf("expected a single argument\n");
    abort();
  }

  t = atoi(argv[1]);

  if (t <= 0) {
    printf("bad thread count: %s\n", argv[1]);
    abort();
  }

  for (t_plus = 1; t_plus < t; t_plus *= 2) { }

  setup();

  single_len = n / t;

  /* Last server waiting_gets leftovers if the array doesn't divide evenly: */
  last_t = t - 1;
  last_len = n - ((t - 1) * single_len);

  nodeval_var = (struct fullempty_t **)malloc(t * sizeof(struct fullempty_t *));
  pval_var = (struct fullempty_t **)malloc(t * sizeof(struct fullempty_t *));
  for (i = 0; i < t;i++) {
    nodeval_var[i] = fullempty_create(sizeof(tally_t));
    pval_var[i] = fullempty_create(sizeof(tally_t));
  }

  c = (pthread_t *)malloc(t * sizeof(pthread_t));

  for (i = 0; i < t; i++) {
    pthread_create(&c[i], NULL, start_server, (void *)i);
  }
  
  for (i = 0; i < t; i++) {
    pthread_join(c[i], NULL);
  }
  
  print_result();

  return 0;
}

static tally_t localTally(int size, data_t *myArray);
static tally_t propagateParentTally(int index);
static void localScanGen(tally_t ptally, int size,
                         data_t *myArray, result_t *myResult);

static void *start_server(void *id)
{
  int index = (int)id;
  int n = ((index == last_t) ? last_len : single_len);
  data_t *myArray;
  result_t *myResult;

  myArray = array + (index * single_len);
  myResult = result + (index * single_len);

  {
    tally_t tally, ptally, child_tally;
    int stride;
      
    tally = localTally(n, myArray);
      
    for (stride = 1; stride < t_plus; stride *= 2) {
      if (index % (2 * stride) == 0) {
        if (index + stride < t) {
          fullempty_put(pval_var[index + stride], sizeof(tally_t), &tally);
          fullempty_get(nodeval_var[index + stride], sizeof(tally_t), &child_tally);
          tally = combine(tally, child_tally);
        }
      } else {
        fullempty_put(nodeval_var[index], sizeof(tally_t), &tally);
        break;
      }
    }

    ptally = propagateParentTally(index);

    localScanGen(ptally, n, myArray, myResult);
  }
  
  return NULL;
}

tally_t localTally(int size, data_t *myArray)
{
  int i;
  tally_t tally = init();
 
  for (i = 0; i < size; i++)
   tally = accum(tally, i, myArray);

  return tally;
}

tally_t propagateParentTally(int index)
{
  tally_t ptally;
  int stride;
  
  if (index == 0)
    ptally = init();
  else {
    /* Parent supplies left total then parent */
    tally_t ptally2;
    fullempty_get(pval_var[index], sizeof(tally_t), &ptally);
    fullempty_get(pval_var[index], sizeof(tally_t), &ptally2);
    ptally = combine(ptally, ptally2);
  }
  
  /* Propagate total from parent to right children */
  for (stride = 1; stride < t_plus; stride *= 2) {
    if (index % (2 * stride) == 0) {
      if (index + stride < t)
        fullempty_put(pval_var[index + stride], sizeof(tally_t), &ptally);
    } else
      break;
  }

  return ptally;
}

void localScanGen(tally_t ptally, int size,
                  data_t *myArray, result_t *myResult)
{
  int i;
  for (i = 0; i < size; i++)
    ptally = scanGen(ptally, i, myArray, myResult);
}

